"""Script to test the acquisition functions."""
from __future__ import print_function
from __future__ import absolute_import

import os
import unittest

from ase.ga.data import DataConnection

from catlearn.api.ase_data_setup import get_unique, get_train
from catlearn.fingerprint.setup import FeatureGenerator
from catlearn.regression import GaussianProcess
from catlearn.regression.acquisition_functions import AcquisitionFunctions

wkdir = os.getcwd()

train_size, test_size = 45, 5


def classifier(atoms):
    """Simple function to classify atoms objects."""
    return atoms.get_chemical_formula()


class TestAcquisition(unittest.TestCase):
    """Test out the various acquisition routines."""

    def get_data(self):
        """Generate features from atoms objects."""
        # Connect database generated by a GA search.
        gadb = DataConnection('{}/data/gadb.db'.format(wkdir))

        # Get all relaxed candidates from the db file.
        print('Getting candidates from the database')
        all_cand = gadb.get_all_relaxed_candidates(use_extinct=False)

        # Setup the test and training datasets.
        testset = get_unique(atoms=all_cand, size=test_size, key='raw_score')

        trainset = get_train(atoms=all_cand, size=train_size,
                             taken=testset['taken'], key='raw_score')

        # Clear out some old saved data.
        for i in trainset['atoms']:
            del i.info['data']['nnmat']

        # Initiate the fingerprint generators with relevant input variables.
        print('Getting the fingerprints')
        f = FeatureGenerator()

        train_features = f.return_vec(
            trainset['atoms'], [f.nearestneighbour_vec])
        test_features = f.return_vec(
            testset['atoms'], [f.nearestneighbour_vec])

        train_targets = []
        for a in trainset['atoms']:
            train_targets.append(a.info['key_value_pairs']['raw_score'])
        test_targets = []
        for a in testset['atoms']:
            test_targets.append(a.info['key_value_pairs']['raw_score'])

        return train_features, train_targets, trainset['atoms'], \
            test_features, test_targets, testset['atoms']

    def test_acquisition(self):
        """Test acquisition functions."""
        train_features, train_targets, train_atoms, test_features, \
            test_targets, test_atoms = self.get_data()
        # Test prediction routine with gaussian kernel.
        kdict = {'k1': {'type': 'gaussian', 'width': 1., 'scaling': 1.}}
        gp = GaussianProcess(
            train_fp=train_features, train_target=train_targets,
            kernel_dict=kdict, regularization=1e-3,
            optimize_hyperparameters=True, scale_data=True)
        pred = gp.predict(
            test_fp=test_features, test_target=test_targets,
            get_validation_error=True, get_training_error=True,
            uncertainty=True)

        print('gaussian prediction (rmse):',
              pred['validation_error']['rmse_average'])

        af = AcquisitionFunctions()
        acq = af.rank(
            targets=train_targets, predictions=pred['prediction'],
            uncertainty=pred['uncertainty'], train_features=train_features,
            test_features=test_features, metrics=[
                'cdf', 'optimistic', 'gaussian', 'UCB', 'EI', 'PI'])
        self.assertTrue(len(acq['cdf']) == len(pred['prediction']))
        self.assertTrue(len(acq['optimistic']) == len(pred['prediction']))
        self.assertTrue(len(acq['gaussian']) == len(pred['prediction']))
        self.assertTrue(len(acq['UCB']) == len(pred['prediction']))
        self.assertTrue(len(acq['EI']) == len(pred['prediction']))
        self.assertTrue(len(acq['PI']) == len(pred['prediction']))

        acq = af.classify(
            classifier, train_atoms, test_atoms, targets=train_targets,
            predictions=pred['prediction'], uncertainty=pred['uncertainty'],
            train_features=train_features, test_features=test_features,
            metrics=['cdf', 'optimistic', 'gaussian', 'UCB', 'EI', 'PI'])
        self.assertTrue(len(acq['cdf']) == len(pred['prediction']))
        self.assertTrue(len(acq['optimistic']) == len(pred['prediction']))
        self.assertTrue(len(acq['gaussian']) == len(pred['prediction']))
        self.assertTrue(len(acq['UCB']) == len(pred['prediction']))
        self.assertTrue(len(acq['EI']) == len(pred['prediction']))
        self.assertTrue(len(acq['PI']) == len(pred['prediction']))


if __name__ == '__main__':
    unittest.main()
